(*This file is used by the regression tests.*)
(* Add LoadPath "../coq/metatheory".*)
Add LoadPath "../coq" as Ott.
(* Add LoadPath "". *)
(* Add LoadPath "../coq/ln_sets". *)
(* Add LoadPath "../examples/tapl". *)
(* Load "../examples/tapl/_ott_coqrc". *)
